import sys

from qtpy.QtWidgets import QMenuBar, QShortcut, QMenu, QPushButton, QApplication, QWidget
from qtpy.QtGui import QKeySequence
from qtpy.QtCore import Qt, QObject, QEvent

# Hack to work around PySide being imported from nowhere:
import qtpy

if "PySide.QtCore" in sys.modules and qtpy.API != "pyside":
    del sys.modules["PySide.QtCore"]

from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager


class DebuggableMenuBar(QMenuBar):
    def __init__(self, *args, **kwargs):
        super(DebuggableMenuBar, self).__init__(*args, **kwargs)

        self.debugshortcut = QShortcut(QKeySequence("Ctrl+Return"), self, self.showDebugMenu, context=Qt.ApplicationShortcut)

        self._debugmenu = QMenu("Debugging")
        self._debugmenu.addAction("Debug widget", self.startDebugging)

        self.mousedebugger = MouseDebugger()

    def showDebugMenu(self):
        self.addMenu(self._debugmenu)

    def startDebugging(self):
        QApplication.instance().installEventFilter(self.mousedebugger)


class MouseDebugger(QObject):
    def eventFilter(self, obj, event):
        # print(event,obj)
        # print(self.sender())
        if event.type() == QEvent.MouseButtonPress:
            print(QApplication.instance().activeWindow().childAt(event.pos()))
            IPythonDebugger(QApplication.instance().activeWindow().childAt(event.pos())).show()
            QApplication.instance().removeEventFilter(self)
            return True
        return False


class IPythonDebugger(RichJupyterWidget):
    def __init__(self, widget: QWidget):
        super(IPythonDebugger, self).__init__()

        # Setup the kernel
        self.kernel_manager = QtInProcessKernelManager()
        self.kernel_manager.start_kernel()
        kernel = self.kernel_manager.kernel
        kernel.gui = "qt"

        # Push QWidget to the console
        kernel.shell.push({"widget": widget})

        self.kernel_client = self.kernel_manager.client()
        self.kernel_client.start_channels()

        # Setup console widget
        def stop():
            self.kernel_client.stop_channels()
            self.kernel_manager.shutdown_kernel()

        self.exit_requested.connect(stop)


if __name__ == "__main__":
    from qtpy.QtWidgets import QApplication, QMainWindow, QLabel

    app = QApplication([])
    window = QMainWindow()
    window.setCentralWidget(QLabel("test"))
    db = DebuggableMenuBar()
    window.setMenuBar(db)
    window.show()

    app.exec_()



from dataclasses import dataclass
from enum import Enum, auto

from qtpy.QtWidgets import QWidget


class Location(Enum):
    Client = auto()
    Core = auto()
    ExecutionEngine = auto()
    AdaptiveEngine = auto()


class ComputeMode(Enum):
    Blocking = auto()
    Threaded = auto()


class RenderMode(Enum):
    Blocking = auto()
    Background = auto()


@dataclass
class Graph:
    name: str
    compute_with: Location = Location.Client
    compute_mode: ComputeMode = ComputeMode.Blocking
    render_mode: RenderMode = RenderMode.Blocking
    widget: QWidget = None

    def compute(self, data: 'Data', *args):
        ...

    def update(self, data: 'Data', *args):
        ...

    def make_widget(self):
        ...

    # @property
    # def widget(self):
    #     return
    #
    # @widget.setter
    # def widget(self, value):
    #     self._widget = value



import logging
from typing import List, Any, Tuple
from os import environ

from qtpy.QtCore import QObject, Signal, Qt
from qtpy.QtGui import QBrush
from pyqtgraph.dockarea import Dock, DockArea
from pyqtgraph.parametertree import ParameterTree, Parameter
from pyqtgraph.parametertree.parameterTypes import GroupParameter
from qtpy.QtWidgets import QFormLayout, QWidget, QListWidget, QListWidgetItem, QPushButton, QLabel, QSpacerItem, QSizePolicy, QStyle, QToolButton, QHBoxLayout, QVBoxLayout
from loguru import logger

from tsuchinoko.core import CoreState, ExceptionResponse
from tsuchinoko.utils import runengine
from tsuchinoko.utils.threads import invoke_as_event, invoke_in_main_thread


class Singleton(type(QObject)):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Display(Dock):
    ...


class LogHandler(logging.Handler):
    colors = {logging.DEBUG: Qt.gray, logging.ERROR: Qt.darkRed, logging.CRITICAL: Qt.red,
              logging.INFO: Qt.white, logging.WARNING: Qt.yellow}

    def __init__(self, log_widget, level=logging.WARNING):
        super(LogHandler, self).__init__(level=level)
        logging.getLogger().addHandler(self)
        self.log_widget = log_widget

        logger.add(logging.getLogger().handlers[-1], level=level)

    # follows same design as vanilla logger emissions
    def emit(self, record, level=logging.INFO, timestamp=None, icon=None, *args):  # We can have icons!
        item = QListWidgetItem(record.getMessage())
        item.setForeground(QBrush(self.colors[record.levelno]))
        item.setToolTip(timestamp)
        self.log_widget.insertItem(0, item)

        while self.log_widget.count() > 100:
            self.log_widget.takeItem(self.log_widget.count() - 1)

    def sink(self, message:str):
        item = QListWidgetItem(message.strip())
        self.log_widget.insertItem(0, item)

        while self.log_widget.count() > 100:
            self.log_widget.takeItem(self.log_widget.count() - 1)


class Log(Display, logging.Handler):
    def __init__(self):
        super(Log, self).__init__('Log', size=(800, 100))

        log = QListWidget()

        self.addWidget(log)
        self.log_handler = LogHandler(log)

    def log_exception(self, ex: Exception):
        logger.error('An exception occurred in the experiment. More info to follow:')
        logger.exception(ex)


class Configuration(Display, metaclass=Singleton):
    sigRequestParameters = Signal()
    sigPushParameter = Signal(list, object)

    def __init__(self):
        super(Configuration, self).__init__('Configuration', size=(300, 500))

        container_widget = QWidget()
        layout = QVBoxLayout()

        self.parameter = None
        self.parameter_tree = ParameterTree()
        layout.addWidget(self.parameter_tree)
        container_widget.setLayout(layout)
        self.addWidget(container_widget)

    def request_parameters(self):
        self.sigRequestParameters.emit()

    def update_parameters(self, state: dict):
        self.parameter = GroupParameter(name='top')
        self.parameter.restoreState(state)
        self.parameter_tree.setParameters(self.parameter, showTop=False)  # required to hide top
        self.parameter.sigTreeStateChanged.connect(self.push_changes)

    def push_changes(self, sender, changes: List[Tuple[Parameter, str, Any]]):
        for change in changes:
            if len(change) == 3:
                param, change, info = change
                if change == 'value':
                    self.sigPushParameter.emit(self.parameter.childPath(param), info)


class StateManager(Display, metaclass=Singleton):
    sigStart = Signal()
    sigStop = Signal()
    sigPause = Signal()

    def __init__(self):
        super(StateManager, self).__init__('Status', size=(300, 50))

        self.stop_button = QToolButton()
        self.start_pause_button = QToolButton()
        self.state_label = QLabel('...')

        self.stop_button.setIcon(self.style().standardIcon(QStyle.SP_MediaStop))
        self.start_pause_button.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))

        self.start_pause_button.clicked.connect(self._start_or_pause)
        self.stop_button.clicked.connect(self.sigStop)

        layout_widget = QWidget()
        layout_widget.setLayout(QHBoxLayout())

        layout_widget.layout().addWidget(self.stop_button)
        layout_widget.layout().addWidget(self.start_pause_button)
        layout_widget.layout().addWidget(self.state_label)

        self.addWidget(layout_widget)

        self.state = CoreState.Connecting

    def update_state(self, state):
        # set state value immediately
        self._state = state
        # defer setter actions until event can be consumed
        invoke_as_event(setattr, self, 'state', state)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        if state in [CoreState.Starting, CoreState.Pausing, CoreState.Restarting, CoreState.Connecting]:
            self.start_pause_button.setDisabled(True)
            self.stop_button.setDisabled(True)
        elif state in [CoreState.Running]:
            self.start_pause_button.setText('Pause')
            self.start_pause_button.setEnabled(True)
            self.start_pause_button.setIcon(self.style().standardIcon(QStyle.SP_MediaPause))
            self.stop_button.setEnabled(True)
        elif state in [CoreState.Paused]:
            self.start_pause_button.setText('Resume')
            self.start_pause_button.setEnabled(True)
            self.start_pause_button.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))
            self.stop_button.setEnabled(True)
        elif state in [CoreState.Inactive]:
            self.start_pause_button.setText('Start')
            self.start_pause_button.setEnabled(True)
            self.start_pause_button.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))
            self.stop_button.setEnabled(False)

        self.state_label.setText(CoreState(state).name)
        self._state = state

    def _start_or_pause(self):
        if self.start_pause_button.text() == 'Pause':
            self.sigPause.emit()
        elif self.start_pause_button.text() in ['Start', 'Resume']:
            self.sigStart.emit()


class GraphManager(Display, metaclass=Singleton):
    def __init__(self):
        self.dock_area = DockArea()

        super(GraphManager, self).__init__('Graphs', hideTitle=True, size=(500, 500), widget=self.dock_area)

        self.graphs = list()

    def set_graphs(self, graphs):
        self.clear()
        self.graphs.clear()
        for graph in graphs:
            self.register_graph(graph)

    def register_graph(self, graph):
        graph.widget = graph.make_widget()
        display = Dock(graph.name, area=self.dock_area, widget=graph.widget)
        graph.display = display
        self.dock_area.addDock(display, position='below')
        self.graphs.append(graph)

    def update_graphs(self, data, last_data_size):
        for graph in self.graphs:
            try:
                graph.update(data, slice(last_data_size, None))
            except Exception as ex:
                logging.exception(ex)

    def clear(self):
        for graph in self.graphs:
            # NOTE: the parent's parent is always a Display instance containing only that graph
            graph.widget.parent().parent().setParent(None)
            graph.widget.parent().parent().close()
            graph.widget.parent().parent().deleteLater()

    def reset(self):
        self.clear()
        self.set_graphs(self.graphs)



import threading
import time
from asyncio import events
from enum import Enum, auto
from queue import Queue

from loguru import logger

from .messages import FullDataRequest, FullDataResponse, PartialDataRequest, PartialDataResponse, StartRequest, UnknownResponse, PauseRequest, StateRequest, GetParametersRequest, SetParameterRequest, GetParametersResponse, SetParameterResponse, StopRequest, StateResponse, MeasureRequest, \
    MeasureResponse, ConnectRequest, ConnectResponse, ExceptionResponse, PushDataRequest, PushDataResponse, GraphsResponse
from ..adaptive import Engine as AdaptiveEngine, Data
from ..execution import Engine as ExecutionEngine
from ..utils.logging import log_time


class CoreState(Enum):
    Connecting = auto()
    Inactive = auto()
    Starting = auto()
    Running = auto()
    Pausing = auto()
    Paused = auto()
    Resuming = auto()
    Stopping = auto()
    Restarting = auto()
    Exiting = auto()


class Core:
    def __init__(self,
                 execution_engine: ExecutionEngine = None,
                 adaptive_engine: AdaptiveEngine = None):
        self.execution_engine = execution_engine
        self.adaptive_engine = adaptive_engine

        self.iteration = 0

        self._state = CoreState.Inactive
        self._exception_queue = Queue()
        self._forced_position_queue = Queue()

        self.data = Data()
        self._graphs = []

        self.experiment_thread = None

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        logger.info(f'Changing core state to {value}')
        self._state = value

    def set_execution_engine(self, engine: ExecutionEngine):
        self.execution_engine = engine

    def set_adaptive_engine(self, engine: AdaptiveEngine):
        self.adaptive_engine = engine

    def main(self, debug=False):
        loop = events.new_event_loop()  # <---- this ensures the current loop is replaced
        try:
            events.set_event_loop(loop)
            loop.set_debug(debug)
            return loop.run_until_complete(self._main())
        finally:
            try:
                # _cancel_all_tasks(loop)
                loop.run_until_complete(loop.shutdown_asyncgens())
            finally:
                events.set_event_loop(None)
                loop.close()

    async def _main(self, min_response_sleep=.1):
        while self.state != CoreState.Exiting:

            if self.state == CoreState.Running:
                pass
                # await sleep(min_response_sleep)  # short-circuit case
            elif self.state == CoreState.Starting:
                if not len(self.data):
                    self.data = Data(dimensionality=self.adaptive_engine.dimensionality)
                self.adaptive_engine.reset()
                self.experiment_thread = threading.Thread(target=self.experiment_loop, args=())  # must hold ref
                self.experiment_thread.start()
                self.state = CoreState.Running

            elif self.state == CoreState.Inactive:
                pass
                # await sleep(min_response_sleep)

            elif self.state == CoreState.Paused:
                pass
                # await sleep(min_response_sleep)

            elif self.state == CoreState.Pausing:
                self.state = CoreState.Paused

            elif self.state == CoreState.Resuming:
                self.state = CoreState.Running

            elif self.state == CoreState.Stopping:
                self.state = CoreState.Inactive
                self.data = Data()
                # await sleep(min_response_sleep)

            await self.notify_clients()

    def experiment_loop(self):
        while self.state != CoreState.Exiting:
            if self.state == CoreState.Running:
                logger.info(f'Iteration: {len(self.data)}')
                try:
                    self.experiment_iteration()
                except Exception as ex:
                    self._exception_queue.put(ex)
                    self.state = CoreState.Pausing
                    logger.exception(ex)
            elif self.state in [CoreState.Stopping, CoreState.Inactive]:
                return
            else:
                time.sleep(.1)

    def experiment_iteration(self):
        with self.data.iteration():
            with log_time('getting position', cumulative_key='getting position'):
                position = tuple(self.execution_engine.get_position() or [0] * self.data.dimensionality)
            if self._forced_position_queue.empty():
                with log_time('getting targets', cumulative_key='getting targets'):
                    targets = self.adaptive_engine.request_targets(position, n=1)
            else:
                targets = [self._forced_position_queue.get()]
            with log_time('updating targets', cumulative_key='updating targets'):
                self.execution_engine.update_targets(targets)
            with log_time('getting measurements', cumulative_key='getting measurements'):
                new_measurements = self.execution_engine.get_measurements()
            if len(new_measurements):
                with log_time('stashing new measurements', cumulative_key='injecting new measurements'):
                    self.data.inject_new(new_measurements)
                with log_time('updating engine with new measurements', cumulative_key='updating engine with new measurements'):
                    self.adaptive_engine.update_measurements(self.data)
                with log_time('updating metrics', cumulative_key='updating metrics'):
                    self.adaptive_engine.update_metrics(self.data)
            with log_time('training', cumulative_key='training'):
                self.adaptive_engine.train()

    async def notify_clients(self):
        ...

    @property
    def graphs(self):
        execution_graphs = getattr(self.execution_engine, 'graphs', []) or []
        adaptive_graphs = getattr(self.adaptive_engine, 'graphs', []) or []
        return execution_graphs + adaptive_graphs + self._graphs

    @graphs.setter
    def graphs(self, graphs):
        raise NotImplementedError('Updating graphs on server not supported yet.')


class ZMQCore(Core):
    def __init__(self, *args, **kwargs):
        super(ZMQCore, self).__init__(*args, **kwargs)
        # self.start_server()
        self.context = None
        self.poller = None

    def start_server(self):
        import zmq
        from zmq.asyncio import Context, Poller
        self.poller = Poller()
        self.context = Context()
        socket = self.context.socket(zmq.REP)
        socket.bind("tcp://*:5555")
        self.poller.register(socket, zmq.POLLIN)

    def respond_FullDataRequest(self, request):
        with self.data.r_lock():
            return FullDataResponse(self.data.as_dict())

    def respond_PartialDataRequest(self, request):
        if self.data and request.iteration <= len(self.data) and self.state == CoreState.Running:
            with self.data.r_lock():
                partial_data = self.data[request.iteration:]
            return PartialDataResponse(partial_data.as_dict(), request.iteration)
        else:
            return StateResponse(self.state)

    def respond_PushDataRequest(self, request):
        self.data = Data(**request.data)
        return PushDataResponse()

    def respond_StartRequest(self, request):
        if self.state == CoreState.Paused:
            self.state = CoreState.Resuming
        elif self.state == CoreState.Inactive:
            self.state = CoreState.Starting
        return StateResponse(self.state)

    def respond_StopRequest(self, request):
        self.state = CoreState.Stopping
        return StateResponse(self.state)

    def respond_PauseRequest(self, request):
        self.state = CoreState.Pausing
        return StateResponse(self.state)

    def respond_StateRequest(self, request):
        if not self._exception_queue.empty():
            return ExceptionResponse(self._exception_queue.get())
        else:
            return StateResponse(self.state)

    def respond_GetParametersRequest(self, request):
        return GetParametersResponse(self.adaptive_engine.parameters.saveState())

    def respond_SetParameterRequest(self, request):
        self.adaptive_engine.parameters.child(*request.child_path).setValue(request.value)
        return SetParameterResponse(True)

    def respond_MeasureRequest(self, request):
        self._forced_position_queue.put(request.position)
        return MeasureResponse(True)

    def respond_ConnectRequest(self, request):
        return ConnectResponse(self.state)

    def respond_PullGraphsRequest(self, request):
        return GraphsResponse(self.graphs)

    def respond_PushGraphsRequest(self, request):
        self.graphs = request.graphs
        return GraphsResponse(self.graphs)

    async def notify_clients(self):
        import zmq
        if not self.poller:
            self.start_server()

        sockets = dict(await self.poller.poll(timeout=.1))
        for socket in sockets:
            try:
                request = await socket.recv_pyobj(zmq.NOBLOCK)
            except zmq.ZMQError as ex:
                logger.exception(ex)
            else:
                if not request:
                    time.sleep(.1)
                    continue

                logger.info(f"Received request: {request}")
                with log_time('preparing response', cumulative_key='preparing response'):
                    responder = getattr(self, f'respond_{request.__class__.__name__}', None)
                    if responder:
                        response = responder(request)
                    else:
                        response = UnknownResponse()

                logger.info(f'Sending response: {response}')
                await socket.send_pyobj(response)

                if isinstance(response, UnknownResponse):
                    logger.exception(ValueError(f'Unknown request received: {request}'))
                    time.sleep(.1)

    def exit_later(self):
        self.state = CoreState.Exiting

    def exit(self):
        self.exit_later()
        self.experiment_thread.join()



from queue import Queue
from typing import Callable, Tuple, List

from loguru import logger

from . import Engine


class SimpleEngine(Engine):
    def __init__(self, measure_func=Callable[[Tuple[float]], Tuple[float]]):
        super(SimpleEngine, self).__init__()

        self.measure_func = measure_func
        self.position = None
        self.targets = Queue()
        self.new_measurements = []

    def update_targets(self,  targets: List[Tuple]):
        with self.targets.mutex:
            self.targets.queue.clear()

        for target in targets:
            self.targets.put(target)

    def get_position(self) -> Tuple:
        return self.position

    def get_measurements(self) -> List[Tuple]:

        while not self.targets.empty():
            self.position = tuple(self.targets.get())
            self.new_measurements.append(self.measure_func(self.position))

        measurements = self.new_measurements
        self.new_measurements = []
        return measurements



from pyqtgraph import ImageView, PlotWidget
from qtpy.QtCore import Signal, QObject, QEvent, Qt
from qtpy.QtWidgets import QAction

# TODO: map imageitems into target coordinate domain


class RequestRelay(QObject):
    sigRequestMeasure = Signal(tuple)


request_relay = RequestRelay()


class ClickRequesterBase:
    def __init__(self, *args, **kwargs):
        super(ClickRequesterBase, self).__init__(*args, **kwargs)

        self.measure_action = QAction('Queue Measurement at Point')
        self.measure_action.triggered.connect(self.emit_measure_request)
        self._scene().contextMenu.append(self.measure_action)
        self._last_mouse_event_pos = None
        self._install_filter()

    def _install_filter(self):
        ...

    def _scene(self):
        ...

    def eventFilter(self, obj, ev):
        if ev.type() == QEvent.Type.MouseButtonPress:
            if ev.button() == Qt.MouseButton.RightButton:
                self._last_mouse_event_pos = ev.pos()

        return False


class ClickRequester(ClickRequesterBase, ImageView):
    def _scene(self):
        return self.scene

    def _install_filter(self):
        self.ui.graphicsView.installEventFilter(self)

    def emit_measure_request(self, *_):
        app_pos = self._last_mouse_event_pos
        # map to local pos
        local_pos = self.view.vb.mapSceneToView(app_pos)
        request_relay.sigRequestMeasure.emit(local_pos.toTuple())


class ClickRequesterPlot(ClickRequesterBase, PlotWidget):
    def _install_filter(self):
        self.installEventFilter(self)

    def _scene(self):
        return self.sceneObj

    def emit_measure_request(self, *_):
        app_pos = self._last_mouse_event_pos
        # map to local pos
        local_pos = self.plotItem.vb.mapSceneToView(app_pos)
        request_relay.sigRequestMeasure.emit(local_pos.toTuple())



# Code coverage tools 'coverage' and 'pytest-cov' don't seem to correctly trace
# code which is inside methods called from within QThreads, see
# https://github.com/nedbat/coveragepy/issues/686
# To mitigate this problem, I use a custom decorator '@coverage_resolve_trace'
# to be hung onto those method definitions. This will prepend the decorated
# method code with 'sys.settrace(threading._trace_hook)' when a code
# coverage test is detected. When no coverage test is detected, it will just
# pass the original method untouched.
import sys
import threading
from functools import wraps

running_coverage = 'coverage' in sys.modules
if running_coverage: print("\nCode coverage test detected\n")


def coverage_resolve_trace(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        if running_coverage: sys.settrace(threading._trace_hook)
        fn(*args, **kwargs)

    return wrapped



from functools import partial

import numpy as np


def transform(cc, k=100, tau=0.5):
    result = 0.5 + 0.5*np.tanh(k*(cc-tau))
    return result


def explore_target(x, gp, N, k=100, tau=0.5):
    mean = gp.posterior_mean(x)["f(x)"]
    cov = gp.posterior_covariance(x)["v(x)"]
    i = len(gp.points)
    if i <= N:
        return cov
    else:
        return transform(mean, k, tau) * cov


explore_target_100 = partial(explore_target, N=100, k=100, tau=.5)


def radical_gradient(x, gp):
    mean_grad = gp.posterior_mean_grad(x)["df/dx"]
    std = np.sqrt(gp.posterior_covariance(x, variance_only=True)["v(x)"])
    res = np.sqrt(np.linalg.norm(mean_grad, axis=1)) * std
    return res



from functools import lru_cache
from typing import Tuple

import numpy as np
from loguru import logger
from pyqtgraph import HistogramLUTWidget, mkBrush, mkPen, PlotItem, PlotWidget, TableWidget
from qtpy.QtWidgets import QWidget, QHBoxLayout

from tsuchinoko.graphics_items.clouditem import CloudItem
from tsuchinoko.graphics_items.indicatoritem import BetterCurveArrow
from tsuchinoko.graphics_items.mixins import ClickRequesterPlot, ClickRequester
from tsuchinoko.graphs import Graph, Location


class Table(Graph):
    def __init__(self, data_keys: Tuple[str] = None, name: str = 'Table'):
        super(Table, self).__init__(name)
        self.data_keys = data_keys or tuple()

    def make_widget(self):
        self.widget = TableWidget(sortable=False)
        return self.widget

    def update(self, data, update_slice: slice):
        # data = data[update_slice]

        with data.r_lock():
            x = data.positions.copy()
            v = data.variances.copy()
            y = data.scores.copy()

            extra_fields = {data_key: data[data_key].copy() for data_key in self.data_keys}

        values = np.array([x, y, v, *extra_fields.values()])
        names = ['Position', 'Value', 'Variance'] + list(extra_fields.keys())

        rows = range(update_slice.start, len(x))
        table = [{name: value[i] for name, value in zip(names, values)} for i in rows]

        if update_slice.start == 0:
            self.widget.setData(table)
        else:
            for row, table_row in zip(rows, table):
                self.widget.setRow(row, list(table_row.values()))


class ImageViewBlend(ClickRequester):
    pass


class Image(Graph):
    def __init__(self, data_key, name: str = None, accumulates: bool = False):
        self.data_key = data_key
        self.accumulates = accumulates
        super(Image, self).__init__(name=name or data_key)

    def make_widget(self):
        graph = PlotItem()
        self.widget = ImageViewBlend(view=graph)
        graph.vb.invertY(False)  # imageview forces invertY; this resets it
        return self.widget

    def update(self, data, update_slice: slice):
        with data.r_lock():
            v = data[self.data_key].copy()
        if self.accumulates:
            raise NotImplemented('Accumulation in Image graphs not implemented yet')
        else:
            if getattr(v, 'ndim', None) in [2, 3]:
                self.widget.imageItem.setImage(v, autoLevels=self.widget.imageItem.image is None)


class Cloud(Graph):
    def __init__(self, data_key, name: str = None, accumulates: bool = True):
        self.data_key = data_key
        self.accumulates = accumulates

        super(Cloud, self).__init__(name=name)

    def make_widget(self):
        graph = ClickRequesterPlot()
        # scatter = ScatterPlotItem(name='scatter', x=[0], y=[0], size=10, pen=mkPen(None), brush=mkBrush(255, 255, 255, 120))
        self.cloud = CloudItem(name='scatter', size=10)
        histlut = HistogramLUTWidget()
        histlut.setImageItem(self.cloud)

        self.widget = QWidget()
        self.widget.setLayout(QHBoxLayout())

        self.widget.layout().addWidget(graph)
        self.widget.layout().addWidget(histlut)

        graph.addItem(self.cloud)

        # Hard-coded to show max
        self.max_arrow = BetterCurveArrow(self.cloud.scatter, brush=mkBrush('r'))
        self.last_arrow = BetterCurveArrow(self.cloud.scatter, brush=mkBrush('w'))
        # text = TextItem()

        return self.widget

    def update(self, data, update_slice: slice):
        # require_clear = False

        with data.r_lock():
            v = data[self.data_key].copy()

            x, y = zip(*data.positions)

        lengths = len(v), len(x), len(y)

        if not np.all(np.array(lengths) == min(lengths)):
            logger.warning(f'Ragged arrays passed to cloud item with lengths (v, x, y): {lengths}')
            x = x[:min(lengths)]
            y = y[:min(lengths)]
            v = v[:min(lengths)]

        if not len(x):
            return

        # c = [255 * i / len(x) for i in range(len(x))]
        max_index = np.argmax(v)
        last_data_size = min(update_slice.start, len(self.cloud.cData))

        if last_data_size == 0:
            action = self.cloud.setData
        elif not self.accumulates:
            action = self.cloud.updateData
        else:
            action = self.cloud.extendData
            x = x[last_data_size + 1:]
            y = y[last_data_size + 1:]
            v = v[last_data_size + 1:]

        action(x=x,
               y=y,
               c=v,
               data=v,
               # size=5,
               hoverable=True,
               # hoverSymbol='s',
               # hoverSize=6,
               hoverPen=mkPen('b', width=2),
               # hoverBrush=mkBrush('g'),
               )
        # scatter.setData(
        #     [{'pos': (xi, yi),
        #       'size': (vi - min(v)) / (max(v) - min(v)) * 20 + 2 if max(v) != min(v) else 20,
        #       'brush': mkBrush(color=mkColor(255, 255, 255)) if i == len(x) - 1 else mkBrush(
        #           color=mkColor(255 - c, c, 0)),
        #       'symbol': '+' if i == len(x) - 1 else 'o'}
        #      for i, (xi, yi, vi, c) in enumerate(zip(x, y, v, c))])

        self.max_arrow.setIndex(max_index)
        self.last_arrow.setIndex(len(self.cloud.cData) - 1)
        # text.setText(f'Max: {v[max_index]:.2f} ({x[max_index]:.2f}, {y[max_index]:.2f})')
        # text.setPos(x[max_index], y[max_index])


class Plot(Graph):
    def __init__(self, data_key, name: str = None, accumulates: bool = False, widget_kwargs=None):
        self.data_key = data_key
        self.accumulates = accumulates
        self.widget_kwargs = widget_kwargs or dict()
        super(Plot, self).__init__(name=name or data_key)

    def make_widget(self):
        self.widget = PlotWidget(**self.widget_kwargs)
        return self.widget

    def update(self, data, update_slice: slice):
        with data.r_lock():
            v = data[self.data_key].copy()
        if self.accumulates:
            self.widget.plot(np.asarray(v), clear=True)
        else:
            self.widget.plot(np.asarray(v), clear=True)


class GPCamVariance(Cloud):
    def __init__(self):
        super(GPCamVariance, self).__init__(data_key='variances', name='Variance')

    def compute(self, data, engine):
        pass  # This is free from gpCAM


class GPCamScore(Cloud):
    def __init__(self):
        super(GPCamScore, self).__init__(data_key='scores', name='Score')

    def compute(self, data, engine):
        pass  # This is free from gpCAM


class GPCamPosteriorCovariance(Image):
    def __init__(self, shape=(50, 50)):
        self.shape = shape
        super(GPCamPosteriorCovariance, self).__init__(data_key='Posterior Covariance')

    def compute(self, data, engine: 'GPCamInProcessEngine'):
        with data.r_lock():  # quickly grab positions within lock before passing to optimizer
            positions = np.asarray(data.positions.copy())

        # compute posterior covariance without lock
        result_dict = engine.optimizer.posterior_covariance(positions)

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = result_dict['S(x)']


class GPCamAcquisitionFunction(Image):
    compute_with = Location.AdaptiveEngine

    def __init__(self, shape=(50, 50)):
        self.shape = shape
        super(GPCamAcquisitionFunction, self).__init__(data_key='Acquisition Function')

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        from tsuchinoko.adaptive.gpCAM_in_process import acquisition_functions  # avoid circular import

        bounds = tuple(tuple(engine.parameters[('bounds', f'axis_{i}_{edge}')]
                             for edge in ['min', 'max'])
                       for i in range(engine.dimensionality))

        grid_positions = image_grid(bounds, self.shape)

        # calculate acquisition function
        acquisition_function_value = engine.optimizer.evaluate_acquisition_function(grid_positions,
                                                                                    acquisition_function=acquisition_functions[engine.parameters['acquisition_function']])

        try:
            acquisition_function_value = acquisition_function_value.reshape(*self.shape)
        except (ValueError, AttributeError):
            acquisition_function_value = np.array([[0]])

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = acquisition_function_value


class GPCamPosteriorMean(Image):
    compute_with = Location.AdaptiveEngine

    def __init__(self, shape=(50, 50)):
        self.shape = shape
        super(GPCamPosteriorMean, self).__init__(data_key='Posterior Mean')

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        bounds = ((engine.parameters[('bounds', f'axis_{i}_{edge}')]
                   for edge in ['min', 'max'])
                  for i in range(engine.dimensionality))

        grid_positions = image_grid(bounds, self.shape)

        # calculate acquisition function
        posterior_mean_value = engine.optimizer.posterior_mean(grid_positions)['f(x)'].reshape(*self.shape)

        # assign to data object with lock
        with data.w_lock():
            data.states['Posterior Mean'] = posterior_mean_value


@lru_cache(maxsize=10)
def image_grid(bounds, shape):
    return np.asarray(np.meshgrid(*(np.linspace(*bound, num=bins) for bins, bound in zip(shape, bounds)))).T.reshape(-1, 2)



import sys
import threading
import time
from functools import wraps

from loguru import logger
from qtpy.QtCore import QTimer, Signal, QThread, QObject, QEvent, QCoreApplication
from qtpy.QtWidgets import QApplication

from tsuchinoko.utils.coverage import coverage_resolve_trace


def log_error(exception: Exception, value=None, tb=None, **kwargs):
    """
    Logs an exception with traceback. All uncaught exceptions get hooked here

    """

    if not value:
        value = exception
    if not tb:
        tb = exception.__traceback__

    if "caller_name" not in kwargs:
        frame = sys._getframe()
        frame = getattr(frame, "f_back", frame) or frame
        kwargs["caller_name"] = frame.f_code.co_name

    caller_name = kwargs.pop('caller_name', '')

    logger.error("\n The following in {caller_name} was handled safely. It is displayed here for debugging.")
    logger.exception(exception)
    # try:
    #     logging.log(logging.ERROR, "\n" + ' '.join(traceback.format_exception(exception, value, tb)), extra={"caller_name": caller_name}, **kwargs)
    # except AttributeError:
    #     logging.log(logging.ERROR, "\n" + ' '.join(traceback.format_exception_only(exception, value)), extra={"caller_name": caller_name}, **kwargs)


show_busy = lambda *_: None
show_ready = lambda *_: None


# Justification for subclassing qthread: https://woboq.com/blog/qthread-you-were-not-doing-so-wrong.html
class QThreadFuture(QThread):
    """
    A future-like QThread, with many conveniences.
    """

    sigCallback = Signal()
    sigFinished = Signal()
    sigExcept = Signal(Exception)

    def __init__(
            self,
            method,
            *args,
            callback_slot=None,
            finished_slot=None,
            except_slot=None,
            interrupt_callable=None,
            # default_exhandle=True,
            # lock=None,
            # threadkey: str = None,
            showBusy=True,
            # keepalive=True,
            # cancelIfRunning=True,
            priority=QThread.InheritPriority,
            timeout=0,
            name=None,
            **kwargs,
    ):
        super(QThreadFuture, self).__init__()

        # Auto-Kill other threads with same threadkey
        # if threadkey and cancelIfRunning:
        #     for thread in manager.threads:
        #         if thread.threadkey == threadkey:
        #             thread.cancel()
        # self.threadkey = threadkey

        self.name = name
        self.callback_slot = callback_slot
        self.except_slot = except_slot
        # if callback_slot: self.sigCallback.connect(callback_slot)
        self.interrupt_callable = interrupt_callable
        if finished_slot:
            self.sigFinished.connect(finished_slot)
        if except_slot:
            self.sigExcept.connect(except_slot)
        if QApplication.instance():
            # QApplication.instance().aboutToQuit.connect(self.quit)
            QApplication.instance().aboutToQuit.connect(self.interrupt)
        self.method = method
        self.args = args
        self.kwargs = kwargs
        self.timeout = timeout  # ms

        self.cancelled = False
        self.exception = None
        self._purge = False
        self.thread = None
        self.priority = priority
        self.showBusy = showBusy

        # if keepalive:
        #     manager.append(self)

    @property
    def done(self):
        return self.isFinished()

    @property
    def running(self):
        return self.isRunning()

    def start(self):
        """
        Starts the thread
        """
        if self.running:
            raise ValueError("Thread could not be started; it is already running.")

        super(QThreadFuture, self).start(self.priority)
        if self.timeout:
            self._timeout_timer = QTimer.singleShot(self.timeout, self.cancel)

    @coverage_resolve_trace
    def run(self, *args, **kwargs):
        """
        Do not call this from the main thread; you're probably looking for start()
        """
        # if self.threadkey:
        #     threading.current_thread().name = self.threadkey

        threading.current_thread().name = self.name
        self.cancelled = False
        self.exception = None
        if self.showBusy:
            show_busy()
        try:
            value = [None]
            runner = self._run(*args, **kwargs)
            while not self.isInterruptionRequested():
                try:
                    value = next(runner)
                except StopIteration as ex:
                    value = ex.value
                    if not isinstance(value, tuple):
                        value = (value,)
                    if isinstance(self, QThreadFutureIterator) and self.callback_slot:
                        self.callback_slot(*value)
                    self._result = value
                    break
                if not isinstance(value, tuple):
                    value = (value,)
                if isinstance(self, QThreadFutureIterator) and self.yield_slot:
                    invoke_in_main_thread(self.yield_slot, *value)
                elif isinstance(self, QThreadFuture) and self.callback_slot:
                    invoke_in_main_thread(self.callback_slot, *value)

        except Exception as ex:
            self.exception = ex
            self.sigExcept.emit(ex)
            logger.error(f"Error in thread: "
                         f'Method: {getattr(self.method, "__name__", "UNKNOWN")}\n'
                         f"Args: {self.args}\n"
                         f"Kwargs: {self.kwargs}", )
            log_error(ex)
        else:
            self.sigFinished.emit()
        finally:
            if self.showBusy:
                show_ready()

            self.quit()
            if QApplication.instance():
                try:
                    # QApplication.instance().aboutToQuit.disconnect(self.quit)
                    QApplication.instance().aboutToQuit.disconnect(self.interrupt)
                # Somehow the application never had its aboutToQuit connected to quit...
                except (TypeError, RuntimeError) as e:
                    # msg.logError(e)
                    ...

    def _run(self, *args, **kwargs):  # Used to generalize to QThreadFutureIterator
        yield self.method(*self.args, **self.kwargs)

    def result(self):
        if not self.running:
            self.start()
        while not self.done and not self.exception:
            time.sleep(0.01)
        if self.exception:
            return self.exception
        return self._result

    def cancel(self):
        self.cancelled = True
        if self.except_slot:
            invoke_in_main_thread(self.except_slot, InterruptedError("Thread cancelled."))
        self.requestInterruption()
        self.quit()
        self.wait()

    def interrupt(self):
        self.requestInterruption()
        if self.interrupt_callable:
            self.interrupt_callable()
        self.wait()


class QThreadFutureIterator(QThreadFuture):
    """
    Same as QThreadFuture, but emits to the callback_slot for every yielded value of a generator
    """

    def __init__(self, *args, yield_slot=None, **kwargs):
        super(QThreadFutureIterator, self).__init__(*args, **kwargs)
        self.yield_slot = yield_slot

    def _run(self, *args, **kwargs):
        return (yield from self.method(*self.args, **self.kwargs))


class InvokeEvent(QEvent):
    """
    Generic callable containing QEvent
    """

    EVENT_TYPE = QEvent.Type(QEvent.registerEventType())

    def __init__(self, fn, *args, **kwargs):
        QEvent.__init__(self, InvokeEvent.EVENT_TYPE)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs


class Invoker(QObject):
    def event(self, event):
        try:
            if hasattr(event.fn, "signal"):  # check if invoking a signal or a callable
                event.fn.emit(*event.args, *event.kwargs.values())
            else:
                event.fn(*event.args, **event.kwargs)
            return True
        except Exception as ex:
            logger.error("QThreadFuture callback could not be invoked.")
            log_error(ex)
        return False


_invoker = Invoker()


def invoke_in_main_thread(fn, *args, force_event=False, **kwargs):
    """
    Invoke a callable in the main thread. Use this for making callbacks to the gui where signals are inconvenient.
    """
    # co_name = sys._getframe().f_back.f_code.co_name
    # msg.logMessage(f"Invoking {fn} in main thread from {co_name}")

    if not force_event and is_main_thread():
        # we're already in the main thread; just do it!
        fn(*args, **kwargs)
    else:
        QCoreApplication.postEvent(_invoker, InvokeEvent(fn, *args, **kwargs))


def invoke_as_event(fn, *args, **kwargs):
    """Invoke a callable as an event in the main thread."""
    # co_name = sys._getframe().f_back.f_code.co_name
    # msg.logMessage(f"Invoking {fn} in main thread from {co_name}")
    invoke_in_main_thread(fn, *args, force_event=True, **kwargs)


def is_main_thread():
    return threading.current_thread() is threading.main_thread()


def method(
        callback_slot=None,
        finished_slot=None,
        except_slot=None,
        # default_exhandle=True,
        # lock=None,
        # threadkey: str = None,
        showBusy=True,
        priority=QThread.InheritPriority,
        # keepalive=True,
        # cancelIfRunning=True,
        timeout=0,
        block=False,
        name=None,
):
    """
    Decorator for functions/methods to run as RunnableMethods on background QT threads
    Use it as any python decorator to decorate a function with @decorator syntax or at runtime:
    decorated_method = threads.method(callback_slot, ...)(method_to_decorate)
    then simply run it: decorated_method(*args, **kwargs)
    Parameters
    ----------
    callback_slot : function
        Function/method to run on a background thread
    finished_slot : QtCore.Slot
        Slot to call with the return value of the function
    except_slot : QtCore.Slot
        Function object (qt slot), slot to receive exception type, instance and traceback object
    default_exhandle : bool
        Flag to use the default exception handle slot. If false it will not be called
    lock : mutex/semaphore
        Simple lock if multiple access needs to be prevented
    Returns
    -------
    wrap_runnable_method : function
        Decorated function/method
    """

    def wrap_runnable_method(func):
        @wraps(func)
        def _runnable_method(*args, **kwargs):
            future = QThreadFuture(
                func,
                *args,
                callback_slot=callback_slot,
                finished_slot=finished_slot,
                except_slot=except_slot,
                # default_exhandle=default_exhandle,
                # lock=lock,
                # threadkey=threadkey,
                showBusy=showBusy,
                priority=priority,
                # keepalive=keepalive,
                # cancelIfRunning=cancelIfRunning,
                timeout=timeout,
                name=name,
                **kwargs,
            )
            future.start()
            if block:
                future.result()
            return future

        return _runnable_method

    return wrap_runnable_method


def iterator(
        yield_slot=None,
        callback_slot=None,
        finished_slot=None,
        interrupt_signal=None,
        except_slot=None,
        # default_exhandle=True,
        # lock=None,
        # threadkey: str = None,
        showBusy=True,
        priority=QThread.InheritPriority,
        name=None,
        # keepalive=True,
):
    """
    Decorator for iterators/generators to run as RunnableIterators on background QT threads
    Use it as any python decorator to decorate a function with @decorator syntax or at runtime:
    decorated_iterator = threads.iterator(callback_slot, ...)(iterator_to_decorate).
    then simply run it: decorated_iterator(*args, **kwargs)

    Parameters
    ----------
    callback_slot : function
        Function/method to run on a background thread
    finished_slot : QtCore.Slot
        Slot to call with the return value of the function
    interrupt_signal : QtCore.Signal
        Signal to break out of iterator loop prematurely
    except_slot : QtCore.Slot
        Function object (qt slot), slot to receive exception type, instance and traceback object
    lock : mutex/semaphore
        Simple lock if multiple access needs to be prevented

    Returns
    -------
    wrap_runnable_iterator : function
        Decorated iterator/generator
    """

    def wrap_runnable_method(func):
        @wraps(func)
        def _runnable_method(*args, **kwargs):
            future = QThreadFutureIterator(
                func,
                *args,
                yield_slot=yield_slot,
                callback_slot=callback_slot,
                finished_slot=finished_slot,
                except_slot=except_slot,
                # default_exhandle=default_exhandle,
                # lock=lock,
                # threadkey=threadkey,
                showBusy=showBusy,
                priority=priority,
                name=name,
                # keepalive=keepalive,
                **kwargs,
            )
            future.start()

        return _runnable_method

    return wrap_runnable_method



import numpy as np

from gpcam.gp_optimizer import  fvGPOptimizer
from .acquisition_functions import explore_target_100, radical_gradient
from .gpCAM_in_process import GPCAMInProcessEngine
from ..graphs.common import GPCamPosteriorCovariance, GPCamAcquisitionFunction, GPCamPosteriorMean, GPCamAverageCovariance, Table

acquisition_functions = {s: s for s in ['variance', 'shannon_ig', 'ucb', 'maximum', 'minimum', 'covariance', 'gradient', 'explore_target_100']}
acquisition_functions['explore_target_100'] = explore_target_100
acquisition_functions['radical_gradient'] = radical_gradient


class FvgpGPCAMInProcessEngine(GPCAMInProcessEngine):
    """
    A multi-task adaptive engine powered by gpCAM: https://gpcam.readthedocs.io/en/latest/
    """

    def __init__(self, dimensionality, output_dim, output_number, parameter_bounds, hyperparameters, hyperparameter_bounds, **kwargs):
        self.kwargs = kwargs
        self.output_dim = output_dim
        self.output_number = output_number
        super(FvgpGPCAMInProcessEngine, self).__init__(dimensionality, parameter_bounds, hyperparameters, hyperparameter_bounds, **kwargs)

        if dimensionality == 2:
            self.graphs = [GPCamPosteriorCovariance(),
                           GPCamAcquisitionFunction(),
                           GPCamPosteriorMean(),
                           GPCamAverageCovariance(),
                           Table()]
        elif dimensionality > 2:
            self.graphs = [GPCamPosteriorCovariance(),
                           GPCamAverageCovariance(),
                           Table()]

    def init_optimizer(self):
        parameter_bounds = np.asarray([[self.parameters[('bounds', f'axis_{i}_{edge}')]
                                        for edge in ['min', 'max']]
                                       for i in range(self.dimensionality)])
        hyperparameters = np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                      for i in range(self.dimensionality + 1)])

        self.optimizer = fvGPOptimizer(self.dimensionality, self.output_dim, self.output_number, parameter_bounds)
        self.optimizer.tell(np.empty((1, self.dimensionality)), np.empty((1, self.output_number)))  # we'll wipe this out later; required for initialization
        self.optimizer.init_fvgp(hyperparameters)

    def _set_hyperparameter(self, parameter, value):
        self.optimizer.gp_initialized = False  # Force re-initialization
        self.optimizer.init_fvgp(np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                           for i in range(self.dimensionality + 1)]))



import click
from pyqtgraph import mkQApp
import ctypes
import os

from ._version import get_versions

__version__ = get_versions()['version']

from .utils import runengine
from . import parameters  # registers parameter types


del get_versions


@click.command()
@click.argument('core_address', required=False, default='localhost')
def launch_client(core_address='localhost'):

    if os.name == 'nt':
        # https://stackoverflow.com/questions/67599432/setting-the-same-icon-as-application-icon-in-task-bar-for-pyqt5-application
        myappid = 'camera.tsuchinoko'  # arbitrary string
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)  # Allows taskbar icon to be shown on windows

    from .widgets.mainwindow import MainWindow
    qapp = mkQApp('Tsuchinoko')

    main_window = MainWindow(core_address)
    main_window.show()

    exit(qapp.exec_())




import numpy as np
from PySide2.QtCore import Qt
from astropy import modeling
from bluesky import Msg, plan_stubs as bps

from tsuchinoko.utils.threads import invoke_in_main_thread


def tune_max_and_fit(
        detectors, signal, motor,
        start, stop, min_step,
        num=10,
        step_factor=3.0,
        snake=False,
        name='primary',
        expected_spot_size=1,
        debug=False,
        sleep=0,
        dark_value=0,
        *, md=None):
    r"""
    plan: tune a motor to the centroid of signal(motor)

    Initially, traverse the range from start to stop with
    the number of points specified.  Repeat with progressively
    smaller step size until the minimum step size is reached.
    Rescans will be centered on the signal centroid
    (for $I(x)$, centroid$= \sum{I}/\sum{x*I}$)
    with original scan range reduced by ``step_factor``.

    Set ``snake=True`` if your positions are reproducible
    moving from either direction.  This will not necessarily
    decrease the number of traversals required to reach convergence.
    Snake motion reduces the total time spent on motion
    to reset the positioner.  For some positioners, such as
    those with hysteresis, snake scanning may not be appropriate.
    For such positioners, always approach the positions from the
    same direction.

    Note:  Ideally the signal has only one peak in the range to
    be scanned.  It is assumed the signal is not polymodal
    between ``start`` and ``stop``.

    Parameters
    ----------
    detectors : Signal
        list of 'readable' objects
    signal : string
        detector field whose output is to maximize
    motor : object
        any 'settable' object (motor, temp controller, etc.)
    start : float
        start of range
    stop : float
        end of range, note: start < stop
    min_step : float
        smallest step size to use.
    num : int, optional
        number of points with each traversal, default = 10
    step_factor : float, optional
        used in calculating new range after each pass

        note: step_factor > 1.0, default = 3
    snake : bool, optional
        if False (default), always scan from start to stop
    md : dict, optional
        metadata

    Examples
    --------
    Find the center of a peak using synthetic hardware.

    >>> from ophyd.sim import SynAxis, SynGauss
    >>> motor = SynAxis(name='motor')
    >>> det = SynGauss(name='det', motor, 'motor',
    ...                center=-1.3, Imax=1e5, sigma=0.05)
    >>> RE(tune_centroid([det], "det", motor, -1.5, -0.5, 0.01, 10))
    """
    if min_step <= 0:
        raise ValueError("min_step must be positive")
    if step_factor <= 1.0:
        raise ValueError("step_factor must be greater than 1.0")
    try:
        motor_name, = motor.hints['fields']
    except (AttributeError, ValueError):
        motor_name = motor.name
    _md = {'detectors': [det.name for det in detectors],
           'motors': [motor.name],
           'plan_args': {'detectors': list(map(repr, detectors)),
                         'motor': repr(motor),
                         'start': start,
                         'stop': stop,
                         'num': num,
                         'min_step': min_step, },
           'plan_name': 'tune_centroid',
           'hints': {},
           }
    _md.update(md or {})
    try:
        dimensions = [(motor.hints['fields'], 'primary')]
    except (AttributeError, KeyError):
        pass
    else:
        _md['hints'].setdefault('dimensions', dimensions)

    low_limit = min(start, stop)
    high_limit = max(start, stop)

    initial_pos = motor.user_readback.get()

    # @bpp.stage_decorator(list(detectors) + [motor])
    def _tune_core(start, stop, num, signal):

        next_pos = start
        step = (stop - start) / (num - 1)
        peak_position = None
        cur_I = None
        sum_I = 0  # for peak centroid calculation, I(x)
        sum_xI = 0

        xs = []
        ys = []

        next_points = list(np.arange(low_limit, high_limit, step))

        while abs(step) >= min_step:
            yield Msg('checkpoint')
            next_pos = next_points.pop()
            yield from bps.mv(motor, next_pos)
            yield from bps.sleep(sleep)
            ret = (yield from bps.trigger_and_read(detectors + [motor], name=name))
            cur_I = ret[signal]['value']
            xs.append(next_pos)
            ys.append(cur_I)
            # sum_I += cur_I
            # position = ret[motor_name]['value']
            # sum_xI += position * cur_I

            if not next_points:

                arg_max = np.argmax(ys)
                if xs[arg_max] == min(xs):
                    next_points = [xs[arg_max] - step]
                elif xs[arg_max] == max(xs):
                    next_points = [xs[arg_max] + step]
                else:
                    step /= 2
                    next_points = [xs[arg_max] - step, xs[arg_max] + step]

        return xs, ys

    xs, ys = np.asarray((yield from _tune_core(start, stop, num, signal)))

    # fit gaussian to measured points

    model = modeling.models.Gaussian1D(amplitude=ys.max(), mean=xs[np.argmax(ys)], stddev=expected_spot_size) + modeling.models.Const1D(dark_value)
    fitter = modeling.fitting.SLSQPLSQFitter()
    fitted_model = fitter(model, xs, ys)
    if debug:  # and (fitted_model.stddev.value <= .1 or fitted_model.stddev.value >= 100):  # disabled; for debugging purposes
        import pyqtgraph as pg
        def plot_poor_fit(xs, ys):
            w = pg.plot(title=f'Scan of {motor_name}')
            w.addLegend()
            sorted_xs = sorted(xs)
            w.plot(sorted_xs, model(sorted_xs), pen='g', name='initial model')
            w.plot(xs, ys, pen=pg.mkPen(color='w', style=Qt.DashLine))
            w.plot(sorted_xs, fitted_model(sorted_xs), pen='r', name='fit')
            w.addItem(pg.ScatterPlotItem(x=xs, y=ys, size=10, pen=pg.mkPen(None), brush=pg.mkBrush('w')), name='measurement')
            w.show()

        invoke_in_main_thread(plot_poor_fit, xs=xs, ys=ys)

    if start < fitted_model.mean_0.value < stop and fitted_model.amplitude_0 > dark_value and fitted_model.stddev_0 < 10 * expected_spot_size:
        yield from bps.mov(motor, fitted_model.mean_0.value)

        return fitted_model

    else:
        # on failed fit, move back where we started
        yield from bps.mov(motor, initial_pos)
        return None



import uuid
from functools import cached_property

import numpy as np
from pyqtgraph.parametertree.parameterTypes import SimpleParameter, GroupParameter, ListParameter

from gpcam.gp_optimizer import GPOptimizer
from . import Engine, Data
from .acquisition_functions import explore_target_100, radical_gradient
from ..graphs.common import GPCamVariance, GPCamPosteriorCovariance, GPCamScore, GPCamAcquisitionFunction, GPCamPosteriorMean, Table
from ..parameters import TrainingParameter

acquisition_functions = {s: s for s in ['variance', 'shannon_ig', 'ucb', 'maximum', 'minimum', 'covariance', 'gradient', 'explore_target_100']}
acquisition_functions['explore_target_100'] = explore_target_100
acquisition_functions['radical_gradient'] = radical_gradient


class GPCAMInProcessEngine(Engine):
    """
    An adaptive engine powered by gpCAM: https://gpcam.readthedocs.io/en/latest/
    """
    default_retrain_globally_at = (20, 50, 100, 400, 1000)
    default_retrain_locally_at = (20, 40, 60, 80, 100, 200, 400, 1000)

    def __init__(self, dimensionality, parameter_bounds, hyperparameters, hyperparameter_bounds, **kwargs):
        self.kwargs = kwargs
        self.dimensionality = dimensionality

        for i in range(dimensionality):
            for j, edge in enumerate(['min', 'max']):
                self.parameters[('bounds', f'axis_{i}_{edge}')] = parameter_bounds[i][j]
        for i in range(dimensionality + 1):
            for j, edge in enumerate(['min', 'max']):
                self.parameters[('hyperparameters', f'hyperparameter_{i}_{edge}')] = hyperparameter_bounds[i][j]
            self.parameters.child('hyperparameters', f'hyperparameter_{i}').setValue(hyperparameters[i], blockSignal=self._set_hyperparameter)

        self.init_optimizer()

        self.optimizer.points = np.array([])
        self.optimizer.values = np.array([])
        self.optimizer.variances = np.array([])

        self._completed_training = {'global': set(),
                                    'local': set()}

        if dimensionality == 2:
            self.graphs = [GPCamVariance(),
                           GPCamScore(),
                           GPCamPosteriorCovariance(),
                           GPCamAcquisitionFunction(),
                           GPCamPosteriorMean(),
                           # GPCamAverageCovariance(),
                           Table()]
        elif dimensionality > 2:
            self.graphs = [GPCamPosteriorCovariance(),
                           # GPCamAverageCovariance(),
                           Table()]

    def init_optimizer(self):
        parameter_bounds = np.asarray([[self.parameters[('bounds', f'axis_{i}_{edge}')]
                                        for edge in ['min', 'max']]
                                       for i in range(self.dimensionality)])
        hyperparameters = np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                      for i in range(self.dimensionality + 1)])

        self.optimizer = GPOptimizer(self.dimensionality, parameter_bounds)
        self.optimizer.tell(np.empty((1, self.dimensionality)), np.empty((1,)), np.empty((1,)))  # we'll wipe this out later; required for initialization
        self.optimizer.init_gp(hyperparameters)

    def reset(self):
        self.init_optimizer()

        self.optimizer.points = np.array([])
        self.optimizer.values = np.array([])
        self.optimizer.variances = np.array([])

    @cached_property
    def parameters(self):
        hyper_parameters = [SimpleParameter(title=f'Hyperparameter #{i + 1}', name=f'hyperparameter_{i}', type='float')
                            for i in range(1 + self.dimensionality)]
        hyper_parameters_bounds = [SimpleParameter(title=f'Hyperparameter #{i + 1} {edge}', name=f'hyperparameter_{i}_{edge}', type='float')
                                   for i in range(1 + self.dimensionality) for edge in ['min', 'max']]
        bounds_parameters = [SimpleParameter(title=f'Axis #{i + 1} {edge}', name=f'axis_{i}_{edge}', type='float')
                             for i in range(self.dimensionality) for edge in ['min', 'max']]
        func_parameters = [ListParameter(title='Method', name='method', limits=['global', 'local', 'hgdl']),
                           ListParameter(title='Acquisition Function', name='acquisition_function', limits=list(acquisition_functions.keys())),
                           SimpleParameter(title='Queue Length', name='n', value=1, type='int'),
                           SimpleParameter(title='Population Size (global only)', name='pop_size', value=20, type='int'),
                           SimpleParameter(title='Tolerance', name='tol', value=1e-6, type='float')]

        global_train_parameter = TrainingParameter(title='Train globally at...', name='global_training', addText='Add', children=[
            SimpleParameter(title='N=', name=str(uuid.uuid4()), value=N, type='int') for N in self.default_retrain_globally_at
        ])
        local_train_parameter = TrainingParameter(title='Train locally at...', name='local_training', addText='Add', children=[
            SimpleParameter(title='N=', name=str(uuid.uuid4()), value=N, type='int') for N in self.default_retrain_locally_at
        ])

        # wireup callback-based parameters
        for param in hyper_parameters:
            param.sigValueChanged.connect(self._set_hyperparameter)

        parameters = func_parameters + [GroupParameter(name='bounds', title='Axis Bounds', children=bounds_parameters),
                                        GroupParameter(name='hyperparameters', title='Hyperparameter Bounds', children=hyper_parameters + hyper_parameters_bounds),
                                        global_train_parameter,
                                        local_train_parameter, ]
        return GroupParameter(name='top', children=parameters)

    def _set_hyperparameter(self, parameter, value):
        self.optimizer.gp_initialized = False  # Force re-initialization
        self.optimizer.init_gp(np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                           for i in range(self.dimensionality + 1)]))

    def update_measurements(self, data: Data):
        with data.r_lock():  # quickly grab values within lock before passing to optimizer
            positions = data.positions.copy()
            scores = data.scores.copy()
            variances = data.variances.copy()
        self.optimizer.tell(positions, scores, variances)

    def update_metrics(self, data: Data):
        for graph in self.graphs:
            graph.compute(data, self)

    def request_targets(self, position, n):
        kwargs = {key: self.parameters[key] for key in ['acquisition_function', 'method', 'pop_size', 'tol']}
        kwargs.update({'bounds': np.asarray([[self.parameters[('bounds', f'axis_{i}_{edge}')]
                                              for edge in ['min', 'max']]
                                             for i in range(self.dimensionality)])})
        return self.optimizer.ask(position, n, acquisition_function=acquisition_functions[kwargs.pop('acquisition_function')], **kwargs)['x']

    def train(self):
        for method in ['global', 'local']:
            train_at = set(child.value() for child in self.parameters.child(f'{method}_training').children())

            for N in train_at:
                if len(self.optimizer.values) > N and N not in self._completed_training[method]:
                    self.optimizer.train(np.asarray([[self.parameters[('hyperparameters', f'hyperparameter_{i}_{edge}')]
                                                      for edge in ['min', 'max']]
                                                     for i in range(self.dimensionality + 1)]),
                                         np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                                     for i in range(self.dimensionality + 1)]), method=method)
                    self._completed_training[method].add(N)
                    break  # only does global training if specified for both



import time
from collections import defaultdict
from contextlib import contextmanager
from typing import Any, Union

from loguru import logger

cumulative_time = defaultdict(lambda: 0)
cumulative_count = defaultdict(lambda: 0)

_application_start_time = time.perf_counter_ns()


@contextmanager
def log_time(*args: Any, level: Union[int, str] = "INFO", cumulative_key: str = "", precision: int = 3) -> None:
    start = time.perf_counter_ns()
    yield
    elapsed_time = time.perf_counter_ns() - start

    if cumulative_key:
        cumulative_time[cumulative_key] += elapsed_time
        cumulative_count[cumulative_key] += 1
        extra_args = (
            f"cumulative elapsed: {cumulative_time[cumulative_key] / 1e6} ms elapsed: {elapsed_time / 1e6} ms avg elapsed: {cumulative_time[cumulative_key] / 1e6 / cumulative_count[cumulative_key]:.{precision}f} ms profile: {cumulative_time[cumulative_key] / (time.perf_counter_ns() - _application_start_time) * 100:.1f}%"
            ,)
    else:
        extra_args = (f"elapsed: {elapsed_time / 1e6} ms elapsed",)

    logger.log(level, ' '.join(args + extra_args))



from functools import cached_property

import numpy as np
from pyqtgraph.parametertree.parameterTypes import SimpleParameter, ListParameter, GroupParameter

from tsuchinoko.adaptive import Engine, Data


class RandomInProcess(Engine):
    dimensionality: int = None

    def __init__(self, dimensionality: int, parameter_bounds, max_targets=None):
        self.dimensionality = dimensionality
        self._max_targets = max_targets
        self._target_count = 0

        for i in range(dimensionality):
            for j, edge in enumerate(['min', 'max']):
                self.parameters[('bounds', f'axis_{i}_{edge}')] = parameter_bounds[i][j]

    @cached_property
    def parameters(self):
        bounds_parameters = [SimpleParameter(title=f'Axis #{i + 1} {edge}', name=f'axis_{i}_{edge}', type='float')
                             for i in range(self.dimensionality) for edge in ['min', 'max']]
        func_parameters = [SimpleParameter(title='Queue Length', name='n', value=1, type='int'),]

        parameters = func_parameters + [GroupParameter(name='bounds', title='Axis Bounds', children=bounds_parameters),]
        return GroupParameter(name='top', children=parameters)

    def update_measurements(self, data: Data):
        ...

    def request_targets(self, position, n, **kwargs):
        if self._max_targets and self._target_count > self._max_targets:
            return []

        bounds = [[self.parameters[('bounds', f'axis_{i}_{edge}')]
                   for edge in ['min', 'max']]
                  for i in range(self.dimensionality)]
        targets = [[np.random.uniform(min_, max_) for min_, max_ in bounds] for i in range(self.parameters['n'])]
        self._target_count += len(targets)
        return targets

    def reset(self):
        ...

    def train(self):
        ...

    def update_metrics(self, data: Data):
        ...



from pyqtgraph import CurvePoint, CurveArrow


class BetterCurvePoint(CurvePoint):
    def event(self, ev):
        try:
            return super(BetterCurvePoint, self).event(ev)
        except IndexError:
            return True


class BetterCurveArrow(BetterCurvePoint, CurveArrow):
    pass



from os import path
from setuptools import setup, find_packages
import sys
import versioneer


# NOTE: This file must remain Python 2 compatible for the foreseeable future,
# to ensure that we error out properly for people with outdated setuptools
# and/or pip.
min_version = (3, 9)
if sys.version_info < min_version:
    error = """
tsuchinoko does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(*(sys.version_info[:2] + min_version))
    sys.exit(error)

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(path.join(here, 'requirements.txt')) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    requirements = [line for line in requirements_file.read().splitlines()
                    if not line.startswith('#')]


setup(
    name='tsuchinoko',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="An adaptive optics alignment tool for ALS beamlines utilizing gpCAM.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Ronald J Pandolfi",
    author_email='ronpandolfi@lbl.gov',
    url='https://github.com/lbl-camera/tsuchinoko',
    python_requires='>={}'.format('.'.join(str(n) for n in min_version)),
    packages=find_packages(exclude=['docs', 'tests', 'examples']),
    entry_points={
        "gui_scripts": [
            'tsuchinoko = tsuchinoko:launch_client',
        ],
    },
    include_package_data=True,
    package_data={
        'tsuchinoko': [
            # When adding files here, remember to update MANIFEST.in as well,
            # or else they will not be included in the distribution on PyPI!
            # 'path/to/data_file',
        ]
    },
    install_requires=requirements,
    license='GPLv3+',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ],
    extras_require={
        "docs": ["sphinx", "sphinx-markdown-tables", "numpydoc", "sphinx_copybutton", "myst_parser", "sphinx_rtd_theme", "sphinx_rtd_dark_mode"],
        "tests": ["pytest", "coverage", "coveralls", "codecov", "pylint", "pytest-qt", "pytest-cov", "pytest-lazy-fixture"],
    },
)



from queue import Queue, Empty
from threading import Thread
import time

from loguru import logger

from . import Engine


class ThreadedInProcessEngine(Engine):
    """
    A simple Execution Engine which performs measurements in a background thread.
    """

    def __init__(self, measure_target, get_position=None):
        # These would normally be on the remote end
        self.exiting = False
        self.targets = Queue()
        self.position_getter = get_position
        if get_position:
            self.position = get_position()
        else:
            self.position = (0, 0)

        self.new_measurements = []
        self.measure_target = measure_target
        self.measure_thread = Thread(target=self.measure_loop)
        self.measure_thread.start()

    def update_targets(self, positions):
        with self.targets.mutex:
            self.targets.queue.clear()

        for position in positions:
            self.targets.put(position)

    def measure_loop(self):
        while not self.exiting:
            logger.critical('looking for target')
            try:
                target = tuple(self.targets.get(timeout=.1))
            except Empty:
                continue
            else:
                self.position = target
                measurement = self.measure_target(target)
                self.new_measurements.append(measurement)

    def get_position(self):
        return self.position or self.position_getter()

    def get_measurements(self):
        measurements = self.new_measurements
        self.new_measurements = []
        return measurements


import time
from collections import defaultdict
from queue import Queue, Empty
from typing import Any, Type, Union

from tsuchinoko.widgets.debugmenubar import DebuggableMenuBar

try:
    from yaml import CLoader as Loader, CDumper as Dumper, dump, load
except ImportError:
    from yaml import Loader, Dumper

import zmq
from zmq.error import ZMQError
import numpy as np
from qtpy.QtGui import QIcon
from loguru import logger
from pyqtgraph import mkBrush, mkPen, HistogramLUTWidget, PlotItem
from pyqtgraph.dockarea import DockArea
from qtmodern.styles import dark
from qtpy.QtWidgets import QMainWindow, QApplication, QHBoxLayout, QWidget, QMenuBar, QAction, QStyle, QFileDialog, QDialog, QMessageBox

from tsuchinoko.assets import path
from tsuchinoko.adaptive import Data
from tsuchinoko.core import CoreState
from tsuchinoko.core.messages import PauseRequest, StartRequest, GetParametersRequest, SetParameterRequest, PartialDataRequest, FullDataRequest, StopRequest, Message, StateRequest, StateResponse, GetParametersResponse, FullDataResponse, PartialDataResponse, MeasureRequest, \
    ConnectRequest, ConnectResponse, PushDataRequest, ExceptionResponse, PullGraphsRequest, GraphsResponse
from tsuchinoko.graphics_items.clouditem import CloudItem
from tsuchinoko.graphics_items.indicatoritem import BetterCurveArrow
from tsuchinoko.graphics_items.mixins import ClickRequester, request_relay, ClickRequesterPlot
from tsuchinoko.utils.threads import QThreadFutureIterator, invoke_as_event
from tsuchinoko.widgets.displays import Log, Configuration, GraphManager, StateManager


class ImageViewBlend(ClickRequester):
    pass


class MainWindow(QMainWindow):
    def __init__(self, core_address='localhost'):
        super(MainWindow, self).__init__()

        menubar = DebuggableMenuBar()
        file_menu = menubar.addMenu("&File")
        file_menu.addAction('&New...', self.new_data)
        open_data_action = QAction(self.style().standardIcon(QStyle.SP_DirOpenIcon), 'Open data...', parent=file_menu)
        open_parameters_action = QAction(self.style().standardIcon(QStyle.SP_DirOpenIcon), 'Open parameters...', parent=file_menu)
        save_data_action = QAction(self.style().standardIcon(QStyle.SP_DialogSaveButton), 'Save data as...', parent=file_menu)
        save_parameters_action = QAction(self.style().standardIcon(QStyle.SP_DialogSaveButton), 'Save parameters as...', parent=file_menu)
        file_menu.addAction(open_data_action)
        file_menu.addAction(open_parameters_action)
        file_menu.addAction(save_data_action)
        file_menu.addAction(save_parameters_action)
        file_menu.addAction('E&xit', self.close)
        self.setMenuBar(menubar)

        save_data_action.triggered.connect(self.save_data)
        open_data_action.triggered.connect(self.open_data)
        save_parameters_action.triggered.connect(self.save_parameters)
        open_parameters_action.triggered.connect(self.open_parameters)

        self.setWindowTitle('Tsuchinoko')
        self.setWindowIcon(QIcon(path('tsuchinoko.png')))
        self.resize(1700, 1000)

        self.log_widget = Log()
        self.configuration_widget = Configuration()
        self.state_manager_widget = StateManager()
        self.graph_manager_widget = GraphManager()

        self.dock_area = DockArea()
        self.setCentralWidget(self.dock_area)

        for position, w, *relaltive_to in [('bottom', self.graph_manager_widget),
                                           ('bottom', self.log_widget, self.graph_manager_widget),
                                           ('right', self.configuration_widget, self.graph_manager_widget),
                                           ('top', self.state_manager_widget, self.configuration_widget),
                                           ]:
            self.dock_area.addDock(w, position, *relaltive_to)

        dark(QApplication.instance())

        self.context = zmq.Context()
        self.socket = None
        self.core_address = core_address
        self.init_socket()

        self.state_manager_widget.sigPause.connect(self.pause)
        self.state_manager_widget.sigStart.connect(self.start)
        self.state_manager_widget.sigStop.connect(self.stop)
        self.configuration_widget.sigPushParameter.connect(self.set_parameter)
        self.configuration_widget.sigRequestParameters.connect(self.request_parameters)
        request_relay.sigRequestMeasure.connect(self.request_measure)

        self.update_thread = QThreadFutureIterator(self.update, finished_slot=self.close_zmq, name='tsuchinoko-update')
        self.update_thread.start()

        self.data: Data = Data()
        self.last_data_size = None
        self.callbacks = defaultdict(list)

        self.subscribe(self.state_manager_widget.update_state, StateResponse)
        self.subscribe(self.state_manager_widget.update_state, ConnectResponse)
        self.subscribe(self.configuration_widget.update_parameters, GetParametersResponse, invoke_as_event=True)
        self.subscribe(self._data_callback, FullDataResponse)
        self.subscribe(self._data_callback, PartialDataResponse)
        self.subscribe(self.refresh_state, ConnectResponse)
        self.subscribe(self.log_widget.log_exception, ExceptionResponse)
        self.subscribe(self.graph_manager_widget.set_graphs, GraphsResponse, invoke_as_event=True)

    def init_socket(self):
        if self.socket:
            logger.debug("Closing socket")
            self.socket.close()

        #  Socket to talk to server
        logger.info("Connecting to core server…")
        self.socket = self.context.socket(zmq.REQ)
        self.socket.setsockopt(zmq.LINGER, 5)
        self.socket.connect(f"tcp://{self.core_address}:5555")
        self.socket.RCVTIMEO = 5000
        self.message_queue = Queue()

    def try_connect(self):
        self.message_queue.put(ConnectRequest())

    def get_state(self):
        self.message_queue.put(StateRequest())

    def pause(self):
        self.message_queue.put(PauseRequest())

    def start(self):
        self.message_queue.put(StartRequest())

    def stop(self):
        self.message_queue.put(StopRequest())

    def request_measure(self, pos):
        self.message_queue.put(MeasureRequest(pos))

    def request_parameters(self):
        self.message_queue.put(GetParametersRequest())

    def set_parameter(self, child_path: str, value: Any):
        if child_path:
            self.message_queue.put(SetParameterRequest(child_path, value))

    def update(self):
        self.data = Data()
        self.last_data_size = 0

        while True:
            yield
            request = None

            if self.state_manager_widget.state == CoreState.Stopping:
                self.last_data_size = 0
                self.data = Data()

            if self.state_manager_widget.state == CoreState.Connecting:
                self.try_connect()

            if self.state_manager_widget.state in [CoreState.Pausing, CoreState.Starting, CoreState.Resuming, CoreState.Resuming, CoreState.Stopping]:
                self.get_state()

            try:
                request = self.message_queue.get(timeout=.2)
            except Empty:
                if self.state_manager_widget.state == CoreState.Running:
                    if self.data:
                        request = PartialDataRequest(len(self.data))
                    else:
                        request = FullDataRequest()

            if not request:
                self.get_state()
                continue
            else:
                logger.info(f'request: {request}')

            try:
                self.socket.send_pyobj(request)
                response = self.socket.recv_pyobj()
            except ZMQError as ex:
                logger.warning(f'Unable to connect to core server at {self.core_address}...')
                time.sleep(1)
                # logger.exception(ex)
                if self.context:
                    self.init_socket()
                self.data = Data()  # wipeout data and get a full update next time
                self.last_data_size = 0
                self.state_manager_widget.update_state(CoreState.Connecting)
            else:
                logger.info(f'response: {response}')
                if not response:
                    self.get_state()
                else:
                    for callback, as_event in self.callbacks[type(response)]:
                        if as_event:
                            invoke_as_event(callback, *response.payload)
                        else:
                            callback(*response.payload)

    def _data_callback(self, data_payload, last_data_size=None):
        if not isinstance(data_payload, dict):  # TODO: Remove when responses are mapped to callbacks
            return

        if last_data_size is not None and last_data_size < len(self.data):
            raise IndexError('Overwriting of previous data prevented.')

        self.data.extend(Data(**data_payload))
        if len(data_payload['positions']):
            invoke_as_event(self.update_graphs, self.data, self.last_data_size)
        self.last_data_size = len(self.data)

    def refresh_state(self, _):
        self.message_queue.queue.clear()
        self.message_queue.put(GetParametersRequest())
        self.message_queue.put(PullGraphsRequest())

    def update_graphs(self, data, last_data_size):
        self.graph_manager_widget.update_graphs(data, last_data_size)
        # x, y = zip(*data.positions)
        #
        # for metric_name in data.metrics:
        #     self._update_graph(metric_name, x, y, data.metrics[metric_name])
        #
        # self._update_graph('score', x, y, data.scores)
        # self._update_graph('variance', x, y, data.variances)

    def subscribe(self, callback, response_type: Union[Type[Message], None] = None, invoke_as_event: bool = False):
        self.callbacks[response_type].append((callback, invoke_as_event))

    def unsubscribe(self, callback, response_type: Union[Type[Message], None] = None):
        self.callbacks[response_type] = list(filter(lambda match_callback, invoke_as_event: match_callback == callback, self.callbacks[response_type]))

    def open_data(self):
        name, filter = QFileDialog.getOpenFileName(filter=("YAML (*.yml)"))
        if not name:
            return

        if len(self.data):
            result = QMessageBox.question(self,
                                          'Clear current data?',
                                          "Loading data will clear the current data set. Would you like to proceed?",
                                          buttons=QMessageBox.StandardButtons(QMessageBox.Yes | QMessageBox.Cancel),
                                          defaultButton=QMessageBox.Yes)
            if result != QMessageBox.Yes:
                return

        self.data = Data(**load(open(name, 'r'), Loader=Loader))
        self.last_data_size = len(self.data)
        self.graph_manager_widget.reset()
        self.update_graphs(self.data, 0)
        if self.state_manager_widget.state == CoreState.Connecting:
            logger.warning('Data has been loaded before connecting to an experiment server. Remember to reload data after a connection is established.')
        else:
            result = QMessageBox.question(self,
                                          'Send data to server?',
                                          "Would you like to send the opened data to the experiment server? This will overwrite the server's current data.",
                                          buttons=QMessageBox.StandardButtons(QMessageBox.Yes | QMessageBox.No),
                                          defaultButton=QMessageBox.Yes)
            if result == QMessageBox.Yes:
                self.message_queue.put(PushDataRequest(self.data.as_dict()))

    def save_data(self):
        name, filter = QFileDialog.getSaveFileName(filter=("YAML (*.yml)"))
        if not name:
            return
        with self.data.r_lock():
            dump(self.data.as_dict(), open(name, 'w'), Dumper=Dumper)

    def open_parameters(self):
        name, filter = QFileDialog.getOpenFileName(filter=("YAML (*.yml)"))
        if not name:
            return
        state = load(open(name, 'r'), Loader=Loader)
        self.configuration_widget.parameter.restoreState(state, addChildren=True, removeChildren=True)

    def save_parameters(self):
        name, filter = QFileDialog.getSaveFileName(filter=("YAML (*.yml)"))
        if not name:
            return
        state = self.configuration_widget.parameter.saveState(filter='user')
        dump(state, open(name, 'w'), Dumper=Dumper)

    def new_data(self):
        if len(self.data):
            result = QMessageBox.question(self,
                                          'Clear current data?',
                                          "There is an active data set in memory. Would you like to proceed with clearing the current data?",
                                          buttons=QMessageBox.StandardButtons(QMessageBox.Yes | QMessageBox.Cancel),
                                          defaultButton=QMessageBox.Yes)
            if result != QMessageBox.Yes:
                return

        self.data = Data()
        self.graph_manager_widget.reset()

    def close_zmq(self):
        if self.update_thread.running:
            logger.info('waiting for update thread to finish')
            self.update_thread.requestInterruption()
            self.update_thread.wait()
        if self.socket:
            logger.debug('Closing socket')
            self.socket.close()
            self.socket = None
        if self.context:
            logger.debug('Closing context')
            self.context.term()
            self.context = None

    def closeEvent(self, event):
        if self.data and len(self.data):
            result = QMessageBox.question(self,
                                      'Save data?',
                                      "You have unsaved data in the active experiment. Do you want to save the data?",
                                      buttons=QMessageBox.StandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel),
                                      defaultButton=QMessageBox.Yes)
            if result == QMessageBox.Yes:
                self.save_data()
            if result in [QMessageBox.Yes, QMessageBox.No]:
                event.accept()
                self.close_zmq()
            else:
                event.ignore()
        else:
            event.accept()
            self.close_zmq()
